@@ -32,16 +32,22 @@
 #endif
 
 #define MOD_NAME		"bt_mod"
-#define BT_VER			"1.0.0"
+#define BT_VER			"1.0.1"
 
-#define BT_FLAG_WARN		(1 << 29)
-#define BT_FLAG_DEBUG		(1 << 30)
+#define BT_FLAG_T_START		(1 << 26)
+#define BT_FLAG_T_STOP		(1 << 27)
+#define BT_FLAG_WARN		(1 << 28)
+#define BT_FLAG_DEBUG		(1 << 29)
+#define BT_FLAG_COMM		(1 << 30)
 #define BT_FLAG_PID		(1 << 31)
 #define	is_bt_record(r)		\
 	(!((r)->flags & (BT_FLAG_WARN|BT_FLAG_DEBUG|BT_FLAG_PID)))
 #define	is_pid_record(r)	((r)->flags & BT_FLAG_PID)
+#define	is_comm_record(r)	((r)->flags & BT_FLAG_COMM)
 #define	is_warn_record(r)	((r)->flags & BT_FLAG_WARN)
 
+#define BT_COMM_LEN		8
+
 struct bt_record {
 	unsigned long	from;
 	unsigned long	to;
@@ -54,12 +60,22 @@ struct pid_record {
 	unsigned long	tv_usec;
 };
 
+struct comm_record {
+	char		comm[BT_COMM_LEN];
+	unsigned long	flags;
+};
+
 struct warn_record {
 	unsigned long	left;
 	unsigned long	reserve;
 	unsigned long	flags;
 };
 
+struct tmr_record {
+	unsigned long long	timestamp;
+	unsigned long		flags;
+};
+
 struct ds_manage {
 	unsigned long	bts_base;
 	unsigned long	bts_index;
@@ -74,8 +90,14 @@ struct ds_manage {
 };
 
 struct pid_manage {
-	pid_t	pid;
-	int	is_wrote;
+	pid_t			pid;
+	char			comm[BT_COMM_LEN];
+	int			is_wrote;
+
+	/* rdtsc value for checking the syscall execution time */
+	long			n_syscall;
+	unsigned long long	syscall_start;
+	unsigned long long	syscall_end;
 };
 
 struct info_per_cpu {
@@ -93,19 +115,25 @@ struct info_per_cpu {
 
 /* MSRs */
 #define MSR_IA32_DS_AREA	0x600
-#define MSR_DEBUGCTLA		0x1d9
-#define MSR_DEBUGCTLA_LBR	(1)
-#define MSR_DEBUGCTLA_BTF	(1 << 1)
-#define MSR_DEBUGCTLA_TR	(1 << 2)
-#define MSR_DEBUGCTLA_BTS	(1 << 3)
-#define MSR_DEBUGCTLA_BTINT	(1 << 4)
-
-#define BITS_PER_BYTE		8
+#define MSR_DEBUGCTL		0x1d9
+struct debugctl_bits {
+	int	lbr;
+	int	tr;
+	int	bts;
+	int	btint;
+};
+#define MSR_DEBUGCTL_P4_BITS	{ 1<<0, 1<<2, 1<<3, 1<<4 }
+#define MSR_DEBUGCTL_PM_BITS	{ 1<<0, 1<<6, 1<<7, 1<<8 }
 
 #define	BTS_BUF_MIN_SIZE	(sizeof(struct bt_record) * 64 * 1024)
 #define	MIN_INT_MARGIN_RECS	(1 * 1024)
 #define	DEFAULT_INT_MARGIN_RECS	(8 * 1024)
 
+/* chk_syscall_time parameter value */
+#define CHK_SCTIME_OFF			0
+#define CHK_SCTIME_ON			1
+#define CHK_SCTIME_ON_WITHOUT_TRACE	2
+
 void bt_enable_per_cpu(void *data);
 void bt_disable_per_cpu(void *data);
 void bt_enable(void);
@@ -115,15 +143,17 @@ void bt_disable(void);
 #define is_start(mode)		((mode) == 1)
 #define is_stop(mode)		((mode) == 2)
 #define is_fr(mode)  		((mode) == 3)
-#define is_apid(mode)  		((mode) == 4)
-#define is_upid(mode)  		((mode) == 5)
+#define is_syscall_pid(mode) 	((mode) == 4)
+#define is_syscall_apid(mode)  	((mode) == 5)
+#define is_upid(mode)  		((mode) == 6)
 
-#define is_user(mode) 		((mode) == 5)
-#define is_kern(mode)		((mode) < 5)
+#define is_user(mode) 		((mode) == 6)
+#define is_kern(mode)		((mode) < 6)
 
 #define is_kern_pid_by_hook(mode)	(is_both(mode) || is_start(mode))
 #define is_kern_all_by_hook(mode)	(is_stop(mode) || is_fr(mode))
-#define is_enable_by_proc(mode)		(is_apid(mode) || is_upid(mode))
+#define is_syscall(mode)	(is_syscall_pid(mode) || is_syscall_apid(mode))
+#define is_enable_by_proc(mode)	(is_syscall(mode) || is_upid(mode))
 
 /* for serial console prints */
 int serial_init(int);
@@ -131,38 +161,8 @@ asmlinkage int serial_prints(const char *, ...);
 
 #else
 
-#define	ALL_PID			-1
-
-struct addr_range {
-	unsigned long		begin;
-	unsigned long		end;
-	struct addr_range	*next;
-};
-
-struct pid_range {
-	pid_t			pid;
-	struct addr_range	*range;
-	struct pid_range	*next;
-};
-
-int alloc_pid_range(pid_t);
-void free_ranges(void);
-char* range2ulongs(char*, unsigned long*, unsigned long*);
-int add_range(pid_t, unsigned long, unsigned long);
-int del_range(pid_t, unsigned long, unsigned long);
-void dump_ranges(void);
-struct pid_range *get_all_ranges(void);
-struct addr_range *get_pid_addr_range(pid_t);
-int is_addr_match(struct bt_record*, struct addr_range*);
-
-#define MAX_LINE_LEN	256
-#define MAX_NAME_LEN	128
-typedef int (*t_func_r2n)(unsigned long, unsigned long, char*, char*, int);
-int open_mmapfile(char*, int*, void**, size_t*);
-void close_mmapfile(int, void*, size_t);
-int is_exists_maps(char*, char*);
-int parse_modules(t_func_r2n, char*);
-int parse_maps(t_func_r2n, char*, char*);
+//#define MAX_LINE_LEN	256
+#define MAX_LINE_LEN	4096
 
 #endif /* __KERNEL__ */
 
