@@ -29,7 +29,8 @@
 #include <asm/apic.h>
 #include "bt.h"
 
-extern struct info_per_cpu bt_info_per_cpu[NR_CPUS];
+extern struct debugctl_bits ctl_bits;
+extern DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 extern unsigned long irq_desc_p;
 extern unsigned long no_irq_type_p;
 
@@ -63,6 +64,7 @@ static void ack(unsigned int irq)
 	int cpu;
 	unsigned int low, high;
 	irq_desc_t *desc;
+	struct info_per_cpu *info;
 
 	desc = (irq_desc_t*)irq_desc_p;
 	spin_unlock(&desc[pmi_irq].lock);
@@ -75,12 +77,13 @@ static void ack(unsigned int irq)
 
 	cpu = smp_processor_id();
 
-	if (!bt_info_per_cpu[cpu].ds_manage) {
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	if (!info->ds_manage) {
 		spin_lock(&desc[pmi_irq].lock);
 		return;
 	}
-	rdmsr(MSR_DEBUGCTLA, low, high);
-	wrmsr(MSR_DEBUGCTLA, low & ~MSR_DEBUGCTLA_TR, high);
+	rdmsr(MSR_DEBUGCTL, low, high);
+	wrmsr(MSR_DEBUGCTL, low & ~ctl_bits.tr, high);
 
 	/* for DEBUG
 	{
@@ -92,7 +95,7 @@ static void ack(unsigned int irq)
 	bts_log_write();
 	irq_mask_clear();
 	spin_lock(&desc[pmi_irq].lock);
-	wrmsr(MSR_DEBUGCTLA, low, high);
+	wrmsr(MSR_DEBUGCTL, low, high);
 }
 
 static void end(unsigned int irq)
