@@ -33,20 +33,29 @@
 #include <linux/vmalloc.h>
 #include <linux/interrupt.h>
 #include <asm/pgtable.h>
+#include <asm/unistd.h>
 #include "bt.h"
 
+struct debugctl_bits ctl_bits;
+struct debugctl_bits ctl_bits_pm = MSR_DEBUGCTL_PM_BITS;
+struct debugctl_bits ctl_bits_p4 = MSR_DEBUGCTL_P4_BITS;
+unsigned int ctl_on_bits, ctl_off_bits;
+
 cpumask_t enable;
 unsigned long *pid_tbl;
-int pid_tbl_ulongs;
+unsigned long *syscall_pid_tbl;
 int pid_max;
-int long_bits;
-spinlock_t pid_tbl_lock = SPIN_LOCK_UNLOCKED;
+////spinlock_t pid_tbl_lock = SPIN_LOCK_UNLOCKED;
+DECLARE_BITMAP(syscall_exe_tbl, NR_syscalls);
+DECLARE_BITMAP(syscall_filter_tbl, NR_syscalls);
 
 unsigned long irq_desc_p;
 unsigned long no_irq_type_p;
 unsigned long switch_to_addr;
 int switch_to_size;
 
+int chk_sctime;
+
 size_t btsbuf_size = BTS_BUF_MIN_SIZE;
 size_t bts_int_margin_recs = DEFAULT_INT_MARGIN_RECS;
 size_t subbuf_size = 2 * 1024 * 1024;
@@ -54,7 +63,7 @@ size_t subbuf_num = 16;
 size_t subbuf_sleep_threshold = 8;
 int mode;
 
-struct info_per_cpu bt_info_per_cpu[NR_CPUS];
+DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 
 /* prototypes */
 int reg_probe(void);
@@ -65,8 +74,9 @@ void proc_cleanup(void);
 static void bt_mod_cleanup(void);
 
 void write_bt_records(void*, size_t);
-void write_pid_record(pid_t);
+void write_pid_record(pid_t, char*);
 void write_warn_record(unsigned long);
+void write_tmr_record(unsigned long long, long);
 int relfs_init(void);
 void relfs_cleanup(void);
 void relfs_flush(void);
@@ -99,16 +109,29 @@ void bts_log_write(void)
 	 */
 	local_irq_save(flags);
 	cpu = smp_processor_id();
-	info = &bt_info_per_cpu[cpu];
+	info = &per_cpu(bt_info_per_cpu, cpu);
 	ds_mng = info->ds_manage;
 	pid_mng = &info->pid_manage;
-	if (!ds_mng || ds_mng->bts_index <= ds_mng->bts_base)
+
+	if (chk_sctime != CHK_SCTIME_ON_WITHOUT_TRACE
+	    && (!ds_mng || ds_mng->bts_index <= ds_mng->bts_base))
 		goto EXIT;
 
 	if (!pid_mng->is_wrote) {
-		write_pid_record(pid_mng->pid);
+		write_pid_record(pid_mng->pid, pid_mng->comm);
 		pid_mng->is_wrote = 1;
 	}
+	if (chk_sctime != CHK_SCTIME_OFF && is_syscall(mode)) {
+		if (pid_mng->syscall_start) {
+			write_tmr_record(pid_mng->syscall_start,
+					 pid_mng->n_syscall);
+			pid_mng->syscall_start = 0;
+		}
+		if (pid_mng->syscall_end) {
+			write_tmr_record(pid_mng->syscall_end, -1);
+			pid_mng->syscall_end = 0;
+		}
+	}
 	write_bt_records((void*)ds_mng->bts_base,
 			 ds_mng->bts_index - ds_mng->bts_base);
 	if (ds_mng->bts_index >= bts_warn_index(ds_mng)) {
@@ -122,92 +145,107 @@ EXIT:
 	local_irq_restore(flags);
 }
 
-int is_trace_pid(pid_t pid)
+int is_trace_pid(pid_t pid, unsigned long *pid_tbl)
 {
 	int rc = 0;
-	unsigned long flags, mask;
+	////unsigned long flags;
 
-	spin_lock_irqsave(&pid_tbl_lock, flags);
-	if (pid > pid_max)
+	////spin_lock_irqsave(&pid_tbl_lock, flags);
+	if (pid <= 0 || pid >= pid_max)
 		rc = 0;
-	else {
-		mask = 1 << (pid % long_bits);
-		rc = !!(pid_tbl[pid / long_bits] & mask);
-	}
-	spin_unlock_irqrestore(&pid_tbl_lock, flags);
+	else
+		rc = test_bit(pid, pid_tbl);
+	////spin_unlock_irqrestore(&pid_tbl_lock, flags);
 	return rc;
 }
 
-void add_pid_tbl(pid_t pid)
+void add_pid_tbl(pid_t pid, unsigned long *pid_tbl)
 {
-	unsigned long flags, mask;
+	////unsigned long flags;
 
-	spin_lock_irqsave(&pid_tbl_lock, flags);
-	if (pid <= pid_max) {
-		mask = 1 << (pid % long_bits);
-		pid_tbl[pid / long_bits] |= mask;
-	}
-	spin_unlock_irqrestore(&pid_tbl_lock, flags);
+	////spin_lock_irqsave(&pid_tbl_lock, flags);
+	if (pid > 0 && pid < pid_max)
+		set_bit(pid, pid_tbl);
+	////spin_unlock_irqrestore(&pid_tbl_lock, flags);
 }
 
-void remove_pid_tbl(pid_t pid)
+void remove_pid_tbl(pid_t pid, unsigned long *pid_tbl)
 {
-	unsigned long flags, mask;
+	////unsigned long flags;
 
-	spin_lock_irqsave(&pid_tbl_lock, flags);
-	if (pid <= pid_max) {
-		mask = 1 << (pid % long_bits);
-		pid_tbl[pid / long_bits] &= ~mask;
-	}
-	spin_unlock_irqrestore(&pid_tbl_lock, flags);
+	////spin_lock_irqsave(&pid_tbl_lock, flags);
+	if (pid > 0 && pid < pid_max)
+		clear_bit(pid, pid_tbl);
+	////spin_unlock_irqrestore(&pid_tbl_lock, flags);
 }
 
-#define MSR_BTS_ON_BITS \
-	(MSR_DEBUGCTLA_BTS | MSR_DEBUGCTLA_TR | MSR_DEBUGCTLA_BTINT)
+int is_trace_syscall(long n_syscall)
+{
+	int rc = 0;
 
-#define MSR_BTS_OFF_BITS \
-	(MSR_DEBUGCTLA_BTS | MSR_DEBUGCTLA_TR | MSR_DEBUGCTLA_BTINT | \
-	 MSR_DEBUGCTLA_LBR)
+	if (n_syscall < 0 || n_syscall >= NR_syscalls)
+		rc = 0;
+	else
+		rc = test_bit(n_syscall, syscall_filter_tbl);
+	return rc;
+}
+
+void add_syscall_tbl(long n_syscall, unsigned long *syscall_tbl)
+{
+
+	if (n_syscall >= 0 && n_syscall < NR_syscalls)
+		set_bit(n_syscall, syscall_tbl);
+}
 
 void bts_facility_on(void)
 {
 	int cpu = smp_processor_id();
 	unsigned int low, high;
+	struct info_per_cpu *info;
 
-	if (!bt_info_per_cpu[cpu].ds_manage)
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	if (!info->ds_manage)
 		return;
 
-	rdmsr(MSR_DEBUGCTLA, low, high);
-	low |= MSR_BTS_ON_BITS;
-	wrmsr(MSR_DEBUGCTLA, low, high);
+	//serial_prints("on(%d)\n", cpu);
+	rdmsr(MSR_DEBUGCTL, low, high);
+	low |= ctl_on_bits;
+	if (chk_sctime != CHK_SCTIME_ON_WITHOUT_TRACE)
+		wrmsr(MSR_DEBUGCTL, low, high);
 }
 
 void bts_facility_off(void)
 {
 	int cpu = smp_processor_id();
 	unsigned int low, high;
+	struct info_per_cpu *info;
 
-	if (!bt_info_per_cpu[cpu].ds_manage)
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	if (!info->ds_manage)
 		return;
 
-	rdmsr(MSR_DEBUGCTLA, low, high);
-	low &= ~MSR_BTS_OFF_BITS;
-	wrmsr(MSR_DEBUGCTLA, low, high);
+	rdmsr(MSR_DEBUGCTL, low, high);
+	low &= ~ctl_off_bits;
+	wrmsr(MSR_DEBUGCTL, low, high);
+	//serial_prints("off(%d)\n", cpu);
 }
 
 void bts_facility_save_and_off(int *is_on)
 {
 	int cpu = smp_processor_id();
 	unsigned int low, high;
+	struct info_per_cpu *info;
 
-	if (!bt_info_per_cpu[cpu].ds_manage)
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	if (!info->ds_manage)
 		return;
 
-	rdmsr(MSR_DEBUGCTLA, low, high);
-	*is_on = low & MSR_DEBUGCTLA_TR;
+	rdmsr(MSR_DEBUGCTL, low, high);
+	*is_on = low & ctl_bits.tr;
 	if (*is_on) {
-		low &= ~MSR_BTS_OFF_BITS;
-		wrmsr(MSR_DEBUGCTLA, low, high);
+		low &= ~ctl_off_bits;
+		wrmsr(MSR_DEBUGCTL, low, high);
+		//serial_prints("off(%d)\n", cpu);
 	}
 }
 
@@ -215,20 +253,40 @@ void bts_facility_restore(int is_on)
 {
 	int cpu = smp_processor_id();
 	unsigned int low, high;
+	struct info_per_cpu *info;
 
-	if (!bt_info_per_cpu[cpu].ds_manage)
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	if (!info->ds_manage)
 		return;
 
-	rdmsr(MSR_DEBUGCTLA, low, high);
-	if ((low ^ is_on) & MSR_DEBUGCTLA_TR) {
-		if (is_on)
-			low |= MSR_BTS_ON_BITS;
-		else
-			low &= ~MSR_BTS_OFF_BITS;
-		wrmsr(MSR_DEBUGCTLA, low, high);
+	rdmsr(MSR_DEBUGCTL, low, high);
+	if ((low ^ is_on) & ctl_bits.tr) {
+		if (is_on) {
+			//serial_prints("on(%d)\n", cpu);
+			low |= ctl_on_bits;
+		} else
+			low &= ~ctl_off_bits;
+		wrmsr(MSR_DEBUGCTL, low, high);
+		/*
+		if (!is_on)
+			serial_prints("off(%d)\n", cpu);
+			*/
 	}
 }
 
+void bts_on_and_set_pid_info(int cpu, struct task_struct *t)
+{
+	struct info_per_cpu *info;
+	struct pid_manage *pid_mng;
+
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	pid_mng = &info->pid_manage;
+	pid_mng->pid = t->pid;
+	memcpy(pid_mng->comm, t->comm, BT_COMM_LEN);
+	pid_mng->is_wrote = 0;
+	bts_facility_on();
+}
+
 void bt_enable_per_cpu(void *data)
 {
 	int cpu;
@@ -241,7 +299,7 @@ void bt_enable_per_cpu(void *data)
 		      smp_processor_id(), preemptible());
 		      */
 	cpu = smp_processor_id();
-	info = &bt_info_per_cpu[cpu];
+	info = &per_cpu(bt_info_per_cpu, cpu);
 	if (!info->ds_manage)
 		return;
 
@@ -249,7 +307,7 @@ void bt_enable_per_cpu(void *data)
 	if (is_kern_pid_by_hook(mode)) {
 		if (!cpu_isset(cpu, enable))
 			return;
-		add_pid_tbl(pid);
+		add_pid_tbl(pid, pid_tbl);
 	} else
 		cpu_set(cpu, enable);
 
@@ -259,12 +317,10 @@ void bt_enable_per_cpu(void *data)
 	if (is_upid(mode))
 		return;
 
-	if (is_kern_all_by_hook(mode) || is_trace_pid(pid)) {
+	if (is_kern_all_by_hook(mode) ||
+	    (!is_syscall_pid(mode) && is_trace_pid(pid, pid_tbl))) {
 		pid_mng = &info->pid_manage;
-		pid_mng->pid = pid;
-		pid_mng->is_wrote = 0;
-
-		bts_facility_on();
+		bts_on_and_set_pid_info(cpu, current);
 		info->on_off_cnt++;
 	}
 }
@@ -280,7 +336,7 @@ void bt_disable_per_cpu(void *from_procfs)
 		      smp_processor_id(), preemptible());
 		      */
 	cpu = smp_processor_id();
-	info = &bt_info_per_cpu[cpu];
+	info = &per_cpu(bt_info_per_cpu, cpu);
 	if (!info->ds_manage)
 		return;
 
@@ -292,11 +348,12 @@ void bt_disable_per_cpu(void *from_procfs)
 	 * processing. It is NULL to stop at once when executed from procfs.
 	 */
 	bts_facility_save_and_off(&save);
-	if (from_procfs || is_kern_all_by_hook(mode) || is_trace_pid(pid)) {
+	if (from_procfs || is_kern_all_by_hook(mode) ||
+	    is_trace_pid(pid, pid_tbl)) {
 		info->on_off_cnt++;
 		bts_log_write();
 		if (is_kern_pid_by_hook(mode))
-			remove_pid_tbl(pid);
+			remove_pid_tbl(pid, pid_tbl);
 	} else
 		bts_facility_restore(save);
 	if (is_both(mode) && !from_procfs)
@@ -323,13 +380,14 @@ EXPORT_SYMBOL_GPL(bt_disable);
 /* export for retrieve trace data from crash dump */
 extern struct rchan *bt_channel;
 extern size_t bt_dropped;
-EXPORT_SYMBOL_GPL(bt_info_per_cpu);
+EXPORT_PER_CPU_SYMBOL_GPL(bt_info_per_cpu);
 EXPORT_SYMBOL_GPL(bt_channel);
 EXPORT_SYMBOL_GPL(bt_dropped);
 
 static int check_bts_availability(void)
 {
 	unsigned int eax, ebx, ecx, edx;
+	unsigned char family, model;
 	
 	cpuid(0x01, &eax, &ebx, &ecx, &edx);
 	if (!(edx & (1 << 21))) {
@@ -337,8 +395,21 @@ static int check_bts_availability(void)
 		       MOD_NAME);
 		return -ENOSYS;
 	}
-	rdmsr(MSR_DEBUGCTLA, eax, edx);
-	if (eax & MSR_DEBUGCTLA_BTS) {
+	family = (eax >> 8) & 0xf;
+	model = (eax >> 4) & 0xf;
+	if (family == 0xf) {
+		family += (eax >> 20) & 0xff;
+		model += ((eax >> 16) & 0xf) << 4;
+	}
+	if (family == 6 && (model == 9 || model == 13)) /* cpu is Pentium-M? */
+		ctl_bits = ctl_bits_pm;
+	else
+		ctl_bits = ctl_bits_p4;
+	ctl_on_bits  = ctl_bits.bts|ctl_bits.tr|ctl_bits.btint;
+	ctl_off_bits = ctl_bits.bts|ctl_bits.tr|ctl_bits.btint|ctl_bits.lbr;
+
+	rdmsr(MSR_DEBUGCTL, eax, edx);
+	if (eax & ctl_bits.bts) {
 		printk("%s: BTS facility is already used.\n", MOD_NAME);
 		return -EBUSY;
 	}
@@ -350,8 +421,10 @@ static void setup_ds_area_per_cpu(void *data)
 	int cpu = smp_processor_id();
 	struct ds_manage *ds_mng;
 	unsigned int low, high;
+	struct info_per_cpu *info;
 
-	ds_mng = bt_info_per_cpu[cpu].ds_manage;
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	ds_mng = info->ds_manage;
 	if (!ds_mng)
 		return;
 	rdmsr(MSR_IA32_DS_AREA, low, high);
@@ -362,6 +435,7 @@ static int setup_ds_area(void)
 {
 	int i;
 	struct ds_manage *ds_mng;
+	struct info_per_cpu *info;
 
 	btsbuf_size -= btsbuf_size % sizeof(struct bt_record); 
 	for (i = 0; i < NR_CPUS; i++) {
@@ -370,7 +444,8 @@ static int setup_ds_area(void)
 
 		ds_mng = (struct ds_manage*)vmalloc(sizeof(struct ds_manage)
 						    + btsbuf_size);
-		bt_info_per_cpu[i].ds_manage = ds_mng;
+		info = &per_cpu(bt_info_per_cpu, i);
+		info->ds_manage = ds_mng;
 		if (!ds_mng) {
 			printk("%s: BTS buffer cannot allocate.\n", MOD_NAME);
 			return -ENOMEM;
@@ -391,7 +466,7 @@ static void cleanup_ds_area(void)
 	struct info_per_cpu *info;
 
 	for (i = 0; i < NR_CPUS; i++) {
-		info = &bt_info_per_cpu[i];
+		info = &per_cpu(bt_info_per_cpu, i);
 		if (info->ds_manage) {
 			vfree(info->ds_manage);
 			info->ds_manage = NULL;
@@ -468,6 +543,8 @@ module_param(no_irq_type_p, ulong, 0);
 module_param(switch_to_addr, ulong, 0);
 module_param(switch_to_size, int, 0);
 
+module_param(chk_sctime, int, 0);
+
 MODULE_AUTHOR("Yumiko Sugita <sugita@sdl.hitachi.co.jp>,\n" \
 	      "\t\tSatoshi Fujiwara <sa-fuji@sdl.hitachi.co.jp>");
 MODULE_DESCRIPTION("Hardware BTS facility controller");
