@@ -36,81 +36,88 @@
 #include <linux/smp.h>
 #include <linux/vmalloc.h>
 #include <linux/mm.h>
+#include <asm/unistd.h>
 #include "bt.h"
 
 struct proc_dir_entry *proc_btrax;
 static struct proc_dir_entry *proc_enable;
 static struct proc_dir_entry *proc_pid;
+static struct proc_dir_entry *proc_filter_syscall;
+static struct proc_dir_entry *proc_exec_syscall;
 static struct proc_dir_entry *proc_btsbuf_sz;
 static struct proc_dir_entry *proc_bts_mrecs;
-extern struct info_per_cpu bt_info_per_cpu[NR_CPUS];
+extern DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 extern cpumask_t enable;
 extern unsigned long *pid_tbl;
-extern int pid_tbl_ulongs;
+extern unsigned long *syscall_pid_tbl;
+extern unsigned long syscall_exe_tbl[];
+extern unsigned long syscall_filter_tbl[];
 extern int pid_max;
-extern int long_bits;
 extern int mode;
 extern size_t btsbuf_size;
 extern size_t bts_int_margin_recs;
 
-extern void bt_enable(void);
-extern void bt_disable(void);
-extern int add_pid_tbl(pid_t);
+#define BT_COLLECT_NAME	"bt_collect_log"
+extern pid_t bt_collect_pid;
 
-static int in_pid_write;
+void bt_enable(void);
+void bt_disable(void);
+void add_pid_tbl(pid_t, unsigned long*);
+void remove_pid_tbl(pid_t, unsigned long*);
+void add_syscall_tbl(int, unsigned long*);
 
-/* proc_pid_read is exists for debug purpose */
-static int proc_pid_read(char *buf, char **start, off_t off,
-			 int count, int *eof, void *data)
+static int proc_read_bitmap(char *buf, char **start, off_t off,
+			    int count, int *eof, void *data,
+			    unsigned long *tbl, int n_max)
 {
-	int i, j, len, max = 10;
-	unsigned long mask;
-	pid_t pid, pid_max_in_tbl = 0;
+	int len, max = 10;
+	int n, n_max_in_tbl = 0;
 	off_t loff;
 	char *fmt, lbuf[max + 1], *p;
 
 	loff = 0;
 	*start = p = buf;
 	
-	/* first, check pid_max_in_tbl */
-	for (i = 0; i < pid_tbl_ulongs; i++) {
-		if (!pid_tbl[i])
-			continue;
-		for (j = 0, mask = 1; j < long_bits; j++, mask <<= 1) {
-			if (!(pid_tbl[i] & mask))
-				continue;
-			pid = i * long_bits + j;
-			if (pid > pid_max_in_tbl)
-				pid_max_in_tbl = pid;
-		}
+	/* first, check n_max_in_tbl */
+	n_max_in_tbl = 0;
+	for (n = -1;;) {
+		if (n < 0)
+			n = find_first_bit(tbl, n_max);
+		else
+			n = find_next_bit(tbl, n_max, n + 1);
+		if (n >= n_max)
+			break;
+		n_max_in_tbl = n;
 	}
 
-	for (i = 0; i < pid_tbl_ulongs; i++) {
-		if (!pid_tbl[i])
-			continue;
-		for (j = 0, mask = 1; j < long_bits; j++, mask <<= 1) {
-			if (!(pid_tbl[i] & mask))
-				continue;
-			pid = i * long_bits + j;
-			fmt = (pid == pid_max_in_tbl) ? "%d\n" : "%d,";
-			len = snprintf(lbuf, max, fmt, pid);
-			if (len < 0 || len >= max)
-				return -EFAULT;
-			if (loff >= off && loff + len <= off + count) {
-				memcpy(p, lbuf, len);
-				p += len;
-			} else if (loff + len > off + count)
-				break;
-			loff += len;
-		}
+	for (n = -1;;) {
+		if (n < 0)
+			n = find_first_bit(tbl, n_max);
+		else
+			n = find_next_bit(tbl, n_max, n + 1);
+		if (n >= n_max)
+			break;
+		fmt = (n == n_max_in_tbl) ? "%d\n" : "%d,";
+		len = snprintf(lbuf, max, fmt, n);
+		if (len < 0 || len >= max)
+			return -EFAULT;
+		if (loff >= off && loff + len <= off + count) {
+			memcpy(p, lbuf, len);
+			p += len;
+		} else if (loff + len > off + count)
+			break;
+		loff += len;
 	}
 	return p - buf;
 }
 
-static int pid_write_from_buf(char *buf, unsigned long len,
-			      unsigned long *consumed)
+typedef void (*func_write_bitmap)(int val, unsigned long *tbl);
+
+static int __proc_write_bitmap(char *buf, unsigned long len,
+			       unsigned long *consumed, unsigned long *tbl,
+			       func_write_bitmap f)
 {
-	long val;
+	int val;
 	char *p, *p_max, *p_end;
 
 	p_max = buf + len;
@@ -118,14 +125,12 @@ static int pid_write_from_buf(char *buf, unsigned long len,
 		val = simple_strtol(p, &p_end, 0);
 		switch (*p_end) {
 		case ',':
-			if (add_pid_tbl(val))
-				return -EINVAL;
+			f(val, tbl);
 			break;
 		case '\n':
 			if (p_max != p_end + 1)
 				return -EINVAL;
-			if (add_pid_tbl(val))
-				return -EINVAL;
+			f(val, tbl);
 			*consumed = p_max - p;
 			return 1;
 		case '\0':
@@ -140,30 +145,30 @@ static int pid_write_from_buf(char *buf, unsigned long len,
 	return 0;
 }
 
+#define WRITE_BUF_SIZE	64
+
 /* The data length (count) might become shorter than the string length
- * of one pid. At this time, it is necessary to process the string
+ * of one element. At this time, it is necessary to process the string
  * because it terminates abnormally when returns 0.
  * Therefore, it is necessary to preserve it in the static variable buf.
  */
-#define PID_WRITE_BUF_SIZE	64
-static int proc_pid_write(struct file *file, const char *user_buf,
-			  unsigned long count, void *data)
+static int in_bitmap_write;
+static int proc_write_bitmap(struct file *file, const char *user_buf,
+			     unsigned long count, void *data,
+			     unsigned long *tbl, int n_max,
+			     func_write_bitmap f)
 {
-	unsigned long max = PID_WRITE_BUF_SIZE, len, consumed;
+	unsigned long max = WRITE_BUF_SIZE, len, consumed;
 	static unsigned long left;
-	static char buf[PID_WRITE_BUF_SIZE + 1];
+	static char buf[WRITE_BUF_SIZE + 1];
 	char *p, *p_max;
 	int rc;
 
-	if (!is_enable_by_proc(mode) || !cpus_empty(enable))
-		return -EFAULT;
-
-	if (!in_pid_write) {
-		memset(pid_tbl, 0,
-		       pid_tbl_ulongs * (sizeof(long) / sizeof(char)));
+	if (!in_bitmap_write) {
+		bitmap_zero(tbl, n_max);
 		left = 0;
 	}
-	in_pid_write = 1;
+	in_bitmap_write = 1;
 
 	for (p = (char*)user_buf, p_max = (char*)user_buf + count;
 	     p < p_max; p += len) {
@@ -173,11 +178,12 @@ static int proc_pid_write(struct file *file, const char *user_buf,
 		if (copy_from_user(buf + left, p, len))
 			return -EFAULT;
 		buf[left + len] = '\0';
-		if ((rc = pid_write_from_buf(buf, left + len, &consumed)) < 0)
+		rc = __proc_write_bitmap(buf, left + len, &consumed, tbl, f);
+		if (rc < 0)
 			return rc;
 		//printk("BUF:%s:%ld:%d:%ld\n", buf, left + len, rc, consumed);
 		if (rc)	/* finished ? */
-			in_pid_write = 0;
+			in_bitmap_write = 0;
 		if (consumed) {
 			len = consumed - left;
 			left = 0;
@@ -188,6 +194,46 @@ static int proc_pid_write(struct file *file, const char *user_buf,
 	return p - user_buf;
 }
 
+static int proc_pid_read(char *buf, char **start, off_t off,
+			 int count, int *eof, void *data)
+{
+	return proc_read_bitmap(buf, start, off, count, eof, data,
+				pid_tbl, pid_max);
+}
+
+static int proc_pid_write(struct file *file, const char *user_buf,
+			  unsigned long count, void *data)
+{
+	if (!is_enable_by_proc(mode) || !cpus_empty(enable))
+		return -EFAULT;
+	return proc_write_bitmap(file, user_buf, count, data,
+				 pid_tbl, pid_max, add_pid_tbl);
+}
+
+static int proc_exec_syscall_read(char *buf, char **start, off_t off,
+				  int count, int *eof, void *data)
+{
+	return proc_read_bitmap(buf, start, off, count, eof, data,
+				syscall_exe_tbl, NR_syscalls);
+}
+
+static int proc_filter_syscall_read(char *buf, char **start, off_t off,
+				    int count, int *eof, void *data)
+{
+	return proc_read_bitmap(buf, start, off, count, eof, data,
+				syscall_filter_tbl, NR_syscalls);
+}
+
+static int proc_filter_syscall_write(struct file *file, const char *user_buf,
+				     unsigned long count, void *data)
+{
+	if (!is_syscall(mode) || !cpus_empty(enable))
+		return -EFAULT;
+	return proc_write_bitmap(file, user_buf, count, data,
+				 syscall_filter_tbl, NR_syscalls,
+				 add_syscall_tbl);
+}
+
 #if 0
 static void otameshi_print(void)
 {
@@ -222,6 +268,24 @@ static int proc_enable_read(char *buf, char **start, off_t off,
 	return n;
 }
 
+static void set_exclusion_pids(void)
+{
+	int i;
+	struct task_struct *t;
+
+	for (i = 1; i < pid_max; i++)
+		add_pid_tbl(i, pid_tbl);
+	write_lock_irq(&tasklist_lock);
+	t = &init_task;
+	for (t = next_task(t); t != &init_task; t = next_task(t)) {
+		if (!bt_collect_pid && strcmp(t->comm, BT_COLLECT_NAME) == 0)
+			bt_collect_pid = t->pid;
+		remove_pid_tbl(t->pid, pid_tbl);
+		//serial_prints("exclusion:%d\n", t->pid);
+	}
+	write_unlock_irq(&tasklist_lock);
+}
+
 static int proc_enable_write(struct file *file, const char *user_buf,
 			     unsigned long count, void *data)
 {
@@ -240,7 +304,10 @@ static int proc_enable_write(struct file *file, const char *user_buf,
 	if (*p_end != '\0')
 		return -EINVAL;
 
+	//printk("BTRAX:enable write(%ld)\n", val);
 	if (val) {
+		if (is_syscall_apid(mode))
+			set_exclusion_pids();
 		//otameshi_print();
 		bt_enable();
 	} else
@@ -264,16 +331,18 @@ static int proc_on_off_cnt_read(char *buf, char **start, off_t off,
 				int count, int *eof, void *data)
 {
 	int n, cpu;
+	struct info_per_cpu *info;
 
 	cpu = (int)data;
-	n = snprintf(buf, count, "%ld\n", bt_info_per_cpu[cpu].on_off_cnt);
+	info = &per_cpu(bt_info_per_cpu, cpu);
+	n = snprintf(buf, count, "%ld\n", info->on_off_cnt);
 
 	return n;
 }
 
 int proc_init(void)
 {
-	int i, bytes_of_long;
+	int i, bytes;
 	char tmp[3 + 4 + 1]; /* cpuXXXX */
 	struct info_per_cpu *info;
 
@@ -281,15 +350,15 @@ int proc_init(void)
 	if (!proc_btrax)
 		return -ENOMEM;
 
-	bytes_of_long = sizeof(long) / sizeof(char);
-	long_bits = BITS_PER_BYTE * bytes_of_long;
-	pid_tbl_ulongs = pid_max / long_bits;
-	if (pid_max % long_bits)
-		pid_tbl_ulongs++;
-	pid_tbl = vmalloc(pid_tbl_ulongs * bytes_of_long);
+	bytes = BITS_TO_LONGS(pid_max) * (sizeof(long) / sizeof(char));
+	pid_tbl = vmalloc(bytes);
 	if (!pid_tbl)
 		return -ENOMEM;
-	memset(pid_tbl, 0, pid_tbl_ulongs * bytes_of_long);
+	bitmap_zero(pid_tbl, pid_max);
+	syscall_pid_tbl = vmalloc(bytes);
+	if (!syscall_pid_tbl)
+		return -ENOMEM;
+	bitmap_zero(syscall_pid_tbl, pid_max);
 
 	proc_pid = create_proc_entry("pid", 0600, proc_btrax);
 	if (!proc_pid)
@@ -297,6 +366,19 @@ int proc_init(void)
 	proc_pid->read_proc = proc_pid_read;
 	proc_pid->write_proc = proc_pid_write;
 
+	bitmap_fill(syscall_filter_tbl, NR_syscalls);
+	proc_filter_syscall = create_proc_entry("filter_syscall", 0400,
+						proc_btrax);
+	if (!proc_filter_syscall)
+		return -ENOMEM;
+	proc_filter_syscall->read_proc = proc_filter_syscall_read;
+	proc_filter_syscall->write_proc = proc_filter_syscall_write;
+
+	proc_exec_syscall = create_proc_entry("exec_syscall", 0400, proc_btrax);
+	if (!proc_exec_syscall)
+		return -ENOMEM;
+	proc_exec_syscall->read_proc = proc_exec_syscall_read;
+
 	proc_enable = create_proc_entry("enable", 0600, proc_btrax);
 	if (!proc_enable)
 		return -ENOMEM;
@@ -314,7 +396,7 @@ int proc_init(void)
 	proc_bts_mrecs->read_proc = proc_bts_mrecs_read;
 
 	for_each_online_cpu(i) {
-		info = &bt_info_per_cpu[i];
+		info = &per_cpu(bt_info_per_cpu, i);
 		snprintf(tmp, sizeof(tmp), "cpu%d", i);
 		info->p_cpuN = proc_mkdir(tmp, proc_btrax);
 		if (!info->p_cpuN)
@@ -335,7 +417,7 @@ void proc_cleanup(void)
 	struct info_per_cpu *info;
 
 	for (i = 0; i < NR_CPUS; i++) {
-		info = &bt_info_per_cpu[i];
+		info = &per_cpu(bt_info_per_cpu, i);
 		if (info->p_on_off_cnt)
 			remove_proc_entry(info->p_on_off_cnt->name,
 					  info->p_cpuN);
@@ -348,6 +430,10 @@ void proc_cleanup(void)
 		vfree(pid_tbl);
 	if (proc_pid)
 		remove_proc_entry(proc_pid->name, proc_btrax);
+	if (proc_filter_syscall)
+		remove_proc_entry(proc_filter_syscall->name, proc_btrax);
+	if (proc_exec_syscall)
+		remove_proc_entry(proc_exec_syscall->name, proc_btrax);
 	if (proc_btsbuf_sz)
 		remove_proc_entry(proc_btsbuf_sz->name, proc_btrax);
 	if (proc_bts_mrecs)
@@ -355,3 +441,4 @@ void proc_cleanup(void)
 	if (proc_btrax)
 		remove_proc_entry(proc_btrax->name, &proc_root);
 }
+
