@@ -42,6 +42,7 @@ struct rchan *bt_channel = NULL;
 size_t bt_dropped;
 
 static DECLARE_WAIT_QUEUE_HEAD(relfs_write_wait);
+static int is_cleanup;
 
 /* produced/consumed control files */
 extern struct proc_dir_entry *proc_btrax;
@@ -50,7 +51,7 @@ static struct proc_dir_entry *proc_n_subbufs;
 static struct proc_dir_entry *proc_subbuf_threshold;
 static struct proc_dir_entry *proc_dropped;
 
-extern struct info_per_cpu bt_info_per_cpu[NR_CPUS];
+extern DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 
 extern size_t subbuf_size;
 extern size_t subbuf_num;
@@ -93,17 +94,21 @@ void write_bt_records(void *p, size_t size)
 	relfs_relay_write(bt_channel, p, size);
 }
 
-void write_pid_record(pid_t pid)
+void write_pid_record(pid_t pid, char *comm)
 {
 	struct pid_record rec;
+	struct comm_record comm_rec;
 	struct timeval tv;
 
 	do_gettimeofday(&tv);
 	rec.pid = pid;
 	rec.tv_sec = tv.tv_sec;
 	rec.tv_usec = tv.tv_usec | BT_FLAG_PID;
-
 	relfs_relay_write(bt_channel, &rec, sizeof(rec));
+
+	memcpy(comm_rec.comm, comm, BT_COMM_LEN);
+	comm_rec.flags = BT_FLAG_COMM;
+	relfs_relay_write(bt_channel, &comm_rec, sizeof(comm_rec));
 }
 
 void write_warn_record(unsigned long left)
@@ -117,6 +122,16 @@ void write_warn_record(unsigned long left)
 	relfs_relay_write(bt_channel, &rec, sizeof(rec));
 }
 
+void write_tmr_record(unsigned long long timestamp, long n_syscall)
+{
+	struct tmr_record rec;
+
+	rec.timestamp = timestamp;
+	rec.flags =
+		n_syscall >= 0 ? BT_FLAG_T_START | n_syscall : BT_FLAG_T_STOP;
+	relfs_relay_write(bt_channel, &rec, sizeof(rec));
+}
+
 size_t get_ready_subbuf_num(struct rchan_buf *buf)
 {
 #if RELAYFS_CHANNEL_VERSION >= 5	// btrax relayfs
@@ -134,22 +149,41 @@ static int is_relfs_writeable(void)
 	size_t ready;
 	int rc, cpu;
 
+	if (is_cleanup)
+		return 1;
 	local_irq_save(flags);
 	cpu = smp_processor_id();
 	buf = bt_channel->buf[cpu];
 	ready = get_ready_subbuf_num(buf);
 	rc = (ready < subbuf_sleep_threshold);
-	/* for DEBUG
-	serial_prints("(%d)is writable: %d(p:%d,c:%d)\n",
-		      cpu, rc, buf->subbufs_produced, buf->subbufs_consumed);
-		      */
+	/* for DEBUG*/
+	if (!rc)
+		serial_prints("(%d)unwritable (p:%d,c:%d)\n", cpu,
+			      buf->subbufs_produced, buf->subbufs_consumed);
+	/**/
 	local_irq_restore(flags);
 	return rc;
 }
 
 void check_and_wait_relfs_write(void)
 {
+#if 0
+	int sleep = 0;
+	int len = sizeof(current->comm);
+	char buf[len + 1];
+
+	if (!is_relfs_writeable()) {
+		sleep = 1;
+		buf[len] = '\0';
+		memcpy(buf, current->comm, len);
+		serial_prints("Zzz (%s)\n", buf);
+	}
+#endif
 	wait_event_interruptible(relfs_write_wait, is_relfs_writeable());
+#if 0
+	if (sleep)
+		serial_prints("!!! (%s)\n", buf);
+#endif
 }
 
 #if RELAYFS_CHANNEL_VERSION < 5
@@ -221,7 +255,7 @@ static int subbuf_start_handler(struct rchan_buf *buf, void *subbuf,
 		if (is_start(mode))
 			return 0;
 		else if (!is_kern_all_by_hook(mode))
-			/* When no-overwrite mode, log data is never dropped. */
+			/* When STOP or FR mode, log data is never dropped. */
 			bt_dropped++;
 	}
 	return 1;
@@ -352,7 +386,7 @@ int relfs_init(void)
 	proc_dropped->read_proc = dropped_read;
 
 	for_each_online_cpu(i) {
-		info = &bt_info_per_cpu[i];
+		info = &per_cpu(bt_info_per_cpu, i);
 		info->p_produced = create_proc_entry("produced", 0400,
 						     info->p_cpuN);
 		if (!info->p_produced)
@@ -446,6 +480,12 @@ void relfs_cleanup(void)
 	int i;
 	struct info_per_cpu *info;
 
+	is_cleanup = 1;
+	wake_up_interruptible_all(&relfs_write_wait);
+	do {
+		yield();
+	} while (waitqueue_active(&relfs_write_wait));
+
 	if (proc_subbuf_sz)
 		remove_proc_entry(proc_subbuf_sz->name, proc_btrax);
 	if (proc_n_subbufs)
@@ -458,7 +498,7 @@ void relfs_cleanup(void)
 	kill_procs_using_relfs();
 
 	for (i = 0; i < NR_CPUS; i++) {
-		info = &bt_info_per_cpu[i];
+		info = &per_cpu(bt_info_per_cpu, i);
 		if (info->p_produced)
 			remove_proc_entry(info->p_produced->name, info->p_cpuN);
 		if (info->p_consumed)
