@@ -24,332 +24,31 @@
 /*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
 /*****************************************************************************/
 
-//#define USE_OLD_REPEAT_CHK
-
-#include "bfd_if.h"
-#include "bt.h"
+#include <getopt.h>
 #include "chk_repeat.h"
-#include <libgen.h>
+#include "bt_ar_parse.h"
+#include "bt_utils.h"
 
 #define BT_EXECPATH_VER	BT_VER
 #define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2006"
 
-#define	chk_next(argc, i)	if ((argc) <= (i)+1) { usage(); err_exit(); }
-
-/* for summary */
-#define MAX_BRI_CACHE	1024
-struct br_info {
-	unsigned int	type;
-	unsigned long	from;
-	unsigned long	next;
-	unsigned long	branch;
-	struct br_info	*link;
+struct proc_each_rec_data {
+	struct addr_range	*range;
+	int			nest_chk;
 };
 
-struct range_to_name {
-	unsigned long	begin;
-	unsigned long	end;
-	unsigned long	offset;
-	char		name[MAX_NAME_LEN];
-	char		*dirname;
-	char		*basename;
-
-	struct bfd_if	bi;
-	struct br_info	*bri_cache;
+struct proc_each_range_data {
+	int	fd;
+	off_t	i_rec_max;
+	off_t	i_start;
+	off_t	i_stop;
+	int	match_found;
+	off_t	match_i_rec;
 };
 
 int output_summary = 0;
 int verbose = 0;
-
-/*----------------------------------------------------------------------------
- *  build range to name structure
- *----------------------------------------------------------------------------
- */
-struct range_to_name **all_r2n;
-int r2n_max = 128;
-
-void dump_r2n(void)
-{
-	int i;
-	struct range_to_name *p;
-
-	if (!all_r2n)
-		return;
-	for (i = 0; i < r2n_max; i++) {
-		p = all_r2n[i];
-		if (p) {
-			printf("====== 0x%08lx:0x%08lx\t%s/%s ======\n",
-			       p->begin, p->end, p->dirname, p->basename);
-			dump_bfd_symbols(&p->bi);
-		}
-	}
-}
-
-void free_r2n(void)
-{
-	int i;
-	struct range_to_name *p;
-
-	if (!all_r2n)
-		return;
-	for (i = 0; i < r2n_max; i++) {
-		p = all_r2n[i];
-		if (p) {
-			//free(p->func_names);
-			free(p);
-		}
-	}
-	free(all_r2n);
-}
-
-struct range_to_name* addr_to_r2n(unsigned long addr)
-{
-	int i;
-	struct range_to_name *p;
-
-	for (i = 0; i < r2n_max; i++) {
-		p = all_r2n[i];
-		if (p && addr >= p->begin && addr <= p->end)
-			return p;
-	}
-	return NULL;
-}
-
-int prepare_obj_file(struct range_to_name *r2n, char *kallsyms)
-{
-	char path[MAX_LINE_LEN];
-
-	snprintf(path, MAX_LINE_LEN, "%s/%s", r2n->dirname, r2n->basename);
-	printf("checking %s...\n", path);
-	if (init_bfd_if(&r2n->bi, strdup(path), kallsyms) < 0)
-		return -1;
-	if (!r2n->begin && !r2n->end) {
-		struct addr_range *range = get_pid_addr_range(ALL_PID);
-		bfd_vma min, max;
-		struct bt_record rec;
-
-		get_min_max_addr(&r2n->bi, &min, &max);
-		rec.from = min;
-		rec.to = max;
-		if (is_addr_match(&rec, range)) {
-			r2n->begin = min;
-			r2n->end = max;
-		} else
-			return 1;
-	} else {
-		r2n->offset = get_offset_addr(&r2n->bi, r2n->begin);
-		/* for DEBUG
-		printf("OFFSET: ");
-		printf_bfd_vma(r2n->offset);
-		printf("\n");
-		*/
-	}
-	return 0;
-}
-
-int is_same_name_registered(char *name)
-{
-	int i, len;
-	struct range_to_name *r2n;
-
-	if (!all_r2n)
-		return 0;
-	for (i = 0; i < r2n_max; i++) {
-		r2n = all_r2n[i];
-		if (!r2n)
-			continue;
-		len = strlen(r2n->dirname);
-		if (strstr(name, r2n->dirname) == name && name[len] == '/' &&
-		    strstr(name, r2n->basename) == name + len + 1) {
-			return 1;
-		}
-	}
-	return 0;
-}
-
-int add_range_to_name(unsigned long begin, unsigned long end, char *name,
-		      char *kallsyms, int is_module)
-{
-	int rc, i, step = 128;
-	struct range_to_name *r2n;
-
-	if (is_same_name_registered(name))
-		return 0;
-	r2n = malloc(sizeof(struct range_to_name));
-	if (!r2n) {
-		fprintf(stderr, "malloc failed.(%s)\n", strerror(errno));
-		return -1;
-	}
-	memset(r2n, 0, sizeof(struct range_to_name));
-	r2n->begin = begin;
-	r2n->end = end;
-	snprintf(r2n->name, MAX_NAME_LEN, "%s", name);
-	r2n->basename = basename(r2n->name);
-	r2n->dirname = dirname(r2n->name);
-
-	for (i = 0; i < r2n_max && all_r2n[i]; i++);
-	if (i >= r2n_max) {
-		all_r2n = realloc(all_r2n, (r2n_max + step) * sizeof(r2n));
-		if (!all_r2n) {
-			free(r2n);
-			fprintf(stderr, "realloc failed.(%s)\n",
-				strerror(errno));
-			return -1;
-		}
-		memset(&all_r2n[i], 0, step * sizeof(r2n));
-		i = r2n_max;
-		r2n_max += step;
-	}
-	all_r2n[i] = r2n;
-	rc = prepare_obj_file(r2n, kallsyms);
-	if (rc < 0)
-		return 0;
-		//return -1;
-	else if (rc > 0) {
-		free(r2n);
-		all_r2n[i] = NULL;
-	}
-	return 0;
-}
-
-#ifdef USE_OLD_REPEAT_CHK
-/*----------------------------------------------------------------------------
- *  repeat checking
- *----------------------------------------------------------------------------
- */
-#define MAX_REPEAT_LEN	2048
-static unsigned long addr_buf[MAX_REPEAT_LEN + 1];
-static unsigned long *abp_top = addr_buf;
-static unsigned long *abp_next = addr_buf;
-#ifndef TEST
-static unsigned long *abp_match;
-static int repeat_cnt;
-#endif
-
-unsigned long* find_abuf(unsigned long addr)
-{
-	unsigned long *p;
-
-	if (abp_top <= abp_next) {
-		for (p = abp_top; p < abp_next; p++)
-			if (*p == addr)
-				return p;
-	} else {
-		for (p = abp_top;
-		     p < addr_buf + MAX_REPEAT_LEN + 1; p++)
-			if (*p == addr)
-				return p;
-		for (p = addr_buf; p < abp_next; p++)
-			if (*p == addr)
-				return p;
-	}
-	return NULL;
-}
-
-int get_repeat_checked_addrs(unsigned long *p, unsigned long *p_max,
-			     unsigned long *p_dest)
-{
-	unsigned long *p_save = p_dest;
-
-	if (p <= p_max) {
-		for (; p < p_max; p++)
-			*p_dest++ = *p;
-	} else {
-		for (; p < addr_buf + MAX_REPEAT_LEN + 1; p++)
-			*p_dest++ = *p;
-		for (p = addr_buf; p < p_max; p++)
-			*p_dest++ = *p;
-	}
-	return (p_dest - p_save);
-}
-
-unsigned long* get_next_abuf_addr(unsigned long *p)
-{
-	return (p == addr_buf + MAX_REPEAT_LEN ? addr_buf : p + 1);
-}
-
-int proc_each_addr(unsigned long, int);
-
-int flush_repeat_chk_buf(void)
-{
-	int i;
-
-	for (i = 0; i < MAX_REPEAT_LEN; i++)
-		if (proc_each_addr(0, 1) < 0)
-			return -1;
-	return 0;
-}
-
-#ifdef TEST
-void chk_repeat(unsigned long addr, unsigned long *p, int *len, int *cnt,
-		int is_flush)
-{
-	*p = addr;
-	*len = 1;
-	*cnt = 0;
-}
-#else
-void chk_repeat(unsigned long addr, unsigned long *p, int *len, int *cnt,
-		int is_flush)
-{
-	unsigned long *p_tmp = NULL;
-
-	*len = 0;
-	*cnt = 0;
-
-	if (abp_match) {
-		if (!is_flush && addr == *abp_match) {
-			abp_match = get_next_abuf_addr(abp_match);
-			if (abp_match == abp_next) {
-				abp_match = abp_top;
-				repeat_cnt++;
-			}
-		} else {
-			if (repeat_cnt)
-				*cnt = repeat_cnt + 1;
-			repeat_cnt = 0;
-			*len = get_repeat_checked_addrs(abp_top, abp_next, p);
-			*abp_match = addr;
-			abp_next = get_next_abuf_addr(abp_match);
-			abp_match = NULL;
-		}
-	} else {
-		if (!is_flush)
-			p_tmp = find_abuf(addr);
-		//printf("(F) %d\n", (p_tmp != NULL));
-		if (p_tmp) {
-			*len = get_repeat_checked_addrs(abp_top, p_tmp, p);
-			abp_top = p_tmp;
-			abp_match = get_next_abuf_addr(abp_top);
-		} else {
-			*abp_next = addr;
-			abp_next = get_next_abuf_addr(abp_next);
-			if (abp_top == abp_next) {
-				abp_top = get_next_abuf_addr(abp_top);
-				*p = *abp_next;
-				*len = 1;
-			}
-		}
-	}
-	/* for debug */
-	/*
-	printf("(D)addr:%08lx ->top:%d, next:%d, match:%d, rcnt:%d\n",
-	       addr, abp_top - addr_buf, abp_next - addr_buf,
-	       abp_match ? abp_match - addr_buf : -1, repeat_cnt);
-	       */
-	/*
-	{
-		int i;
-		printf("(B)                ");
-		p = addr_buf;
-		for (i = 0; i < MAX_REPEAT_LEN + 1; i++)
-			printf("%08lx,", *p++);
-		printf("\n");
-	}
-	*/
-}
-#endif
-#endif
+struct r2n_info r2n_info = { NULL, 128, 0 };
 
 /*----------------------------------------------------------------------------
  *  get function name and mnemonic
@@ -395,59 +94,6 @@ int print_exec_line(unsigned long addr, struct range_to_name *r2n)
  *  normal output support
  *----------------------------------------------------------------------------
  */
-#ifdef USE_OLD_REPEAT_CHK
-int proc_each_addr(unsigned long addr, int is_flush)
-{
-	struct range_to_name *r2n;
-	static unsigned long last_addr = 0;
-	unsigned long abuf[MAX_REPEAT_LEN], *p;
-	int i, len, cnt, indent;
-
-	if (!is_flush && addr == last_addr)
-		return 0;
-	last_addr = addr;
-
-	chk_repeat(addr, abuf, &len, &cnt, is_flush);
-	if (cnt) {
-		printf("===> repeat %d times\n", cnt);
-		indent = 1;
-	} else {
-		indent = 0;
-	}
-	for (p = abuf; p < abuf + len; p++) {
-		r2n = addr_to_r2n(*p);
-		if (indent) {
-			for (i = 0; i < indent; i++)
-				printf("  ");
-		}
-		if (!r2n) {
-			if (*p != 0)
-				printf("--------\t0x%08lx\n", *p);
-			continue;
-		}
-		if (print_exec_line(*p, r2n) < 0)
-			return -1;
-	}
-	return 0;
-}
-
-int proc_each_record_for_normal(struct bt_record *p, struct addr_range *range)
-{
-	if (is_warn_record(p)) {
-		struct warn_record *warn = (struct warn_record*)p;
-		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
-		return 0;
-	}
-	if (is_pid_record(p) || !is_addr_match(p, range))
-		return 0;
-
-	if (proc_each_addr(p->from, 0) < 0)
-		return -1;
-	if (proc_each_addr(p->to, 0) < 0)
-		return -1;
-	return 0;
-}
-#else
 int __comp_N(const void *p1, const void *p2)
 {
 	return ((unsigned long)p1 == (unsigned long)p2);
@@ -467,7 +113,7 @@ void __print_data_N(int nest, const void *p)
 	unsigned long addr = (unsigned long)p;
 
 	__print_nest_space_N(nest);
-	r2n = addr_to_r2n(addr);
+	r2n = addr_to_r2n(&r2n_info, addr);
 	if (!r2n) {
 		if (addr != 0)
 			printf("--------\t0x%08lx\n", addr);
@@ -501,23 +147,31 @@ int proc_each_addr(unsigned long addr)
 	return chk_repeat_each((const void*)addr);
 }
 
-int proc_each_record_for_normal(struct bt_record *p, struct addr_range *range)
+int proc_each_record_for_normal(struct bt_record *__p, off_t i_rec, void *tmp)
 {
-	if (is_warn_record(p)) {
-		struct warn_record *warn = (struct warn_record*)p;
+	struct proc_each_rec_data *data = (struct proc_each_rec_data*)tmp;
+	struct bt_record p = *__p;
+	int rc;
+
+	if (is_warn_record(&p)) {
+		struct warn_record *warn = (struct warn_record*)&p;
 		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
 		return 0;
 	}
-	if (is_pid_record(p) || !is_addr_match(p, range))
+	if (is_pid_record(&p))
+		return 0;
+	rc = chk_fix_from_cache(&r2n_info, &p.from, p.to, NULL, NULL, NULL);
+	if (rc < 0)
+		return rc;
+	if (!is_addr_range_match(p.from, p.to, data->range))
 		return 0;
 
-	if (proc_each_addr(p->from) < 0)
+	if (proc_each_addr(p.from) < 0)
 		return -1;
-	if (proc_each_addr(p->to) < 0)
+	if (proc_each_addr(p.to) < 0)
 		return -1;
 	return 0;
 }
-#endif
 
 /*----------------------------------------------------------------------------
  *  summary output support
@@ -576,47 +230,6 @@ int get_nest_from_ret_addr(unsigned long addr)
 }
 #endif
 
-struct br_info* chk_bri_cache(struct range_to_name *r2n, unsigned long from)
-{
-	int i;
-	struct br_info *p, *p_prev;
-
-	for (i = 0, p_prev = NULL, p = r2n->bri_cache; p;
-	     i++, p_prev = p, p = p->link) {
-		if (p->from == from) {
-			if (p_prev) {
-				p_prev->link = p->link;
-				p->link = r2n->bri_cache;
-				r2n->bri_cache = p;
-			}
-			/* else case means found element is already cache top.*/
-			return p;
-		}
-	}
-	if (i >= MAX_BRI_CACHE) {
-		/* free last element */
-		for (p = r2n->bri_cache; p && p->link && p->link->link;
-		     p = p->link);
-		free(p->link);
-		p->link = NULL;
-	}
-	/* add new element */
-	p = malloc(sizeof(struct br_info));
-	if (!p) {
-		fprintf(stderr, "malloc failed(%s).\n", strerror(errno));
-		return NULL;
-	}
-	p->from = from;
-	if (get_branch_info(&r2n->bi, p->from, &p->type, &p->next, &p->branch,
-			    r2n->offset) < 0) {
-		fprintf(stderr, "get_branch_info failed(0x%08lx).\n", p->from);
-		return NULL;
-	}
-	p->link = r2n->bri_cache;
-	r2n->bri_cache = p;
-	return p;
-}
-
 struct summary {
 	int type;
 	int nest;
@@ -648,7 +261,7 @@ void __print_each_addr_S(unsigned long addr)
 	const char *func_name;
 	struct range_to_name *r2n;
 
-	r2n = addr_to_r2n(addr);
+	r2n = addr_to_r2n(&r2n_info, addr);
 	if (!r2n) {
 		//printf("--------:0x%08lx", addr);
 		printf("0x%08lx", addr);
@@ -686,7 +299,9 @@ void __print_data_S(int nest, const void *p)
 	switch (s->type) {
 	case BTYPE_JMP:  c = 'J'; break;
 	case BTYPE_CALL: c = 'C'; break;
-	case BTYPE_INT:  c = 'I'; break;
+	case BTYPE_INT:
+	case BTYPE_BREAK:
+			 c = 'I'; break;
 	default:         c = '?'; break;
 	}
 	printf("%c ", c);
@@ -726,13 +341,13 @@ int proc_enter(int type, int nest_chk, struct bt_record *p, unsigned long ret)
 	case BTYPE_JMP:
 	case BTYPE_CALL:
 	case BTYPE_INT:
+	case BTYPE_BREAK:
+		/*
+		printf("ENTER: %d, from:%08lx, to:%08lx, ret:%08lx\n",
+		       type, p->from, p->to, ret);
+		       */
 		if (!nest_chk) {
-			s = calloc(1, sizeof(struct summary));
-			if (!s) {
-				fprintf(stderr, "calloc failed.(%s)\n",
-					strerror(errno));
-				return -1;
-			}
+			s = xcalloc(1, sizeof(struct summary));
 			s->type = type;
 			s->nest = nest;
 			s->from = p->from;
@@ -760,8 +375,12 @@ int proc_enter(int type, int nest_chk, struct bt_record *p, unsigned long ret)
 
 void proc_leave(int type, int nest_chk, struct bt_record *p)
 {
-	if (type != BTYPE_CALL && type != BTYPE_INT)
+	if (type != BTYPE_CALL && type != BTYPE_INT && type != BTYPE_BREAK)
 		return;
+	/*
+	printf("LEAVE: %d, from:%08lx, to:%08lx\n",
+	       type, p->from, p->to);
+	       */
 #ifdef RET_CHK
 	/* for DEBUG
 	if (!nest_chk)
@@ -775,74 +394,81 @@ void proc_leave(int type, int nest_chk, struct bt_record *p)
 		min_nest = nest;
 }
 
-int proc_each_record_for_summary(struct bt_record *p, struct addr_range *range,
-				 int nest_chk)
+int proc_each_record_for_summary(struct bt_record *__p, off_t i_rec, void *tmp)
 {
+	struct bt_record p = *__p;
+	struct proc_each_rec_data *data = (struct proc_each_rec_data*)tmp;
 	struct range_to_name *r2n, *r2n_to;
-	struct br_info *bri;
+	int nest_chk = data->nest_chk, rc, type;
+	struct path *p_from;
 
-	if (is_warn_record(p)) {
-		struct warn_record *warn = (struct warn_record*)p;
+	if (!nest_chk && is_warn_record(&p)) {
+		struct warn_record *warn = (struct warn_record*)&p;
 		fprintf(stderr, "WARN: bts left only: %ld\n", warn->left);
 		return 0;
 	}
-	if (is_pid_record(p) || !is_addr_match(p, range))
+	if (is_pid_record(&p))
 		return 0;
-
-	r2n = addr_to_r2n(p->from);
-	r2n_to = addr_to_r2n(p->to);
-	if (!r2n) {
+	r2n_to = addr_to_r2n(&r2n_info, p.to);
+	rc = chk_fix_from_cache(&r2n_info, &p.from, p.to, &r2n, &p_from, NULL);
+	if (rc < 0)
+		return rc;
+	/*
+	if (!is_addr_range_match(p.from, p.to, data->range))
+		return 0;
+		*/
+	if (rc == 0) {
 		if (!r2n_to)
 			return 0;
 		/* iret */
-		proc_leave(BTYPE_INT, nest_chk, p);
+		proc_leave(BTYPE_INT, nest_chk, &p);
 		return 0;
 	}
 	if (!r2n_to) {
 		/* int */
-		return proc_enter(BTYPE_INT, nest_chk, p, p->from);
+		return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
 	}
-	if (!(bri = chk_bri_cache(r2n, p->from)))
-		return -1;
+	type = p_from->type;
+	if (p_from->base != p.from)
+		type = BTYPE_OTHER;
 	/* for DEBUG
 	if (!nest_chk)
 		printf("branch info(%d) from:%08lx branch:%08lx next:%08lx\n",
-		       type, p->from, bri->branch, bri->next);
+		       type, p.from, p_from->jmp, p_from->next);
 		       */
-	switch (bri->type) {
+	switch (type) {
 	case BTYPE_BRANCH:
 	case BTYPE_JMP:
 	case BTYPE_CALL:
-		if (bri->branch == UNKNOWN_BADDR || bri->branch == p->to) {
+		if (p_from->jmp == UNKNOWN_BADDR || p_from->jmp == p.to) {
 			/* branch, jmp, call */
-			return proc_enter(bri->type, nest_chk, p, bri->next);
+			return proc_enter(type, nest_chk, &p,
+					  p_from->next);
 		} else {
 			/* int */
-			return proc_enter(BTYPE_INT, nest_chk, p, p->from);
+			return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
 		}
 	case BTYPE_INT:
+	case BTYPE_BREAK:
 	case BTYPE_OTHER:
 		/* int */
-		return proc_enter(BTYPE_INT, nest_chk, p, p->from);
+		return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
 	case BTYPE_RET:
-		proc_leave(BTYPE_CALL, nest_chk, p);
+		proc_leave(BTYPE_CALL, nest_chk, &p);
 		break;
 	case BTYPE_IRET:
-		proc_leave(BTYPE_INT, nest_chk, p);
+		proc_leave(BTYPE_INT, nest_chk, &p);
 		break;
 	}
 	return 0;
 }
 
-int chk_nest(struct bt_record *p, size_t size, struct addr_range *range)
+int chk_nest(int fd, off_t from, off_t to, struct proc_each_rec_data *data)
 {
-	struct bt_record *p_max;
-
 	nest = 0;
-	for (p_max = (struct bt_record*)((char*)p + size); p < p_max; p++) {
-		if (proc_each_record_for_summary(p, range, 1) < 0)
-			return -1;
-	}
+	if (for_each_block_record(fd, from, to,
+				  proc_each_record_for_summary, data) < 0)
+		return -1;
 	////printf("checked nest1: %d\n", nest);
 	nest = 0 - min_nest;
 #ifdef RET_CHK
@@ -852,83 +478,213 @@ int chk_nest(struct bt_record *p, size_t size, struct addr_range *range)
 	return 0;
 }
 
+/* search address option */
+int search;
+char *search_sym;
+unsigned long search_addr;
+off_t before;
+off_t after;
+off_t display_from;
+off_t display_to;
+/* for debug (hidden options: -W) */
+int search_warn;
+
 /*----------------------------------------------------------------------------
  *  program core
  *----------------------------------------------------------------------------
  */
-int proc_block_records(struct bt_record *p, struct bt_record *p_max) {
-	size_t size = (char*)p_max - (char*)p;
-	struct addr_range *range;
+/* Don't check p->from cause Pentium-M's p->from is not correct. */
+#define __is_search_record(p)	\
+	(is_bt_record(p) && (p)->to == search_addr)
 
-	if (p_max < p) {
-		fprintf(stderr, "from recnum is bigger than to recnum\n");
+static int is_search_record(struct bt_record *p, off_t i_rec)
+{
+	if (search_warn)
+		return is_warn_record(p);
+	else if (display_from)
+		return (i_rec == display_from);
+	else
+		return __is_search_record(p);
+}
+
+int __get_block_range(struct bt_record *p, off_t i_rec, void *tmp)
+{
+	struct proc_each_range_data *data = (struct proc_each_range_data*)tmp;
+
+	if (!is_search_record(p, i_rec))
+		return 0;
+	data->match_found = 1;
+	data->match_i_rec = i_rec + 1;
+	//printf("FOUND  :%lld:%lld\n", i_rec, data->match_i_rec);
+	return 1;
+}
+
+int get_block_range(struct bt_record *p, off_t i_rec, void *tmp)
+{
+	struct proc_each_range_data *data = (struct proc_each_range_data*)tmp;
+	off_t i_tmp;
+
+	data->match_found = 0;
+	data->match_i_rec = -1;
+
+	if (!is_search_record(p, i_rec))
+		return 0;
+	if (data->i_start < 0) {
+		data->i_start = i_rec - before;
+		if (data->i_start < 0)
+			data->i_start = 0;
+	}
+	data->match_found = 1;
+	data->match_i_rec = i_rec + 1;
+	//printf("FOUND>>:%lld:%lld\n", i_rec, data->match_i_rec);
+
+	if (display_to) {
+		data->i_stop = display_to;
+		if (data->i_stop > data->i_rec_max || data->i_stop < 0)
+			data->i_stop = data->i_rec_max;
+		return 1;
+	}
+
+	/* Check whether 'before' or 'after' range contains the search
+	 * address.
+	 */
+	while (data->match_found) {
+		i_tmp = data->match_i_rec + (after > before ? after : before);
+		if (i_tmp > data->i_rec_max) {
+			i_tmp = data->i_rec_max;
+			break;
+		}
+		data->match_found = 0;
+		for_each_block_record(data->fd, data->match_i_rec, i_tmp,
+				      __get_block_range, data);
+	}
+	//printf("B:%lld, A:%lld, M:%lld\n", before, after, data->match_i_rec);
+	data->i_stop = data->match_i_rec + after;
+	if (data->i_stop > data->i_rec_max)
+		data->i_stop = data->i_rec_max;
+	return 1;
+}
+
+int proc_block_records(int fd, off_t from, off_t to)
+{
+	struct proc_each_rec_data data = { NULL, 0 };
+
+	if (to < from) {
+		fprintf(stderr, "from-recnum is bigger than to-recnum\n");
 		return -1;
 	}
-	range = get_pid_addr_range(ALL_PID);
+	data.range = get_pid_addr_range(ALL_PID);
 	if (output_summary) {
-		if (chk_nest(p, size, range) < 0)
+		data.nest_chk = 1;
+		if (chk_nest(fd, from, to, &data) < 0)
 			return -1;
+		data.nest_chk = 0;
 		chk_repeat_start(&chk_repeat_funcs_S);
-		for (; p < p_max; p++) {
-			if (proc_each_record_for_summary(p, range, 0) < 0)
-				return -1;
-		}
+		if (for_each_block_record(fd, from, to,
+					  proc_each_record_for_summary,
+					  &data) < 0)
+			return -1;
 		chk_repeat_end();
 	} else {
-#ifdef USE_OLD_REPEAT_CHK
-		for (; p < p_max; p++) {
-			if (proc_each_record_for_normal(p, range) < 0)
-				return -1;
-		}
-		if (flush_repeat_chk_buf() < 0)
-			return -1;
-#else
 		chk_repeat_start(&chk_repeat_funcs_N);
-		for (; p < p_max; p++) {
-			if (proc_each_record_for_normal(p, range) < 0)
-				return -1;
-		}
+		if (for_each_block_record(fd, from, to,
+					  proc_each_record_for_normal,
+					  &data) < 0)
+			return -1;
 		chk_repeat_end();
-#endif
 	}
 	return 0;
 }
 
+/* Return value:  1 = Find Next (Continue)
+ *                0 = Not Found
+ *               -1 = Error
+ */
+int proc_search_range(off_t i_rec, void *tmp)
+{
+	struct proc_each_range_data *data = (struct proc_each_range_data*)tmp;
+
+	data->i_start = data->i_stop = -1;
+	if (for_each_block_record(data->fd, i_rec, data->i_rec_max,
+				  get_block_range, data) < 0)
+		return -1;
+	if (data->i_start < 0 || data->i_stop < 0)
+		return 0;
+	printf("======== records from %lld to %lld ========\n",
+	       data->i_start, data->i_stop);
+	if (proc_block_records(data->fd, data->i_start, data->i_stop) < 0)
+		return -1;
+	printf("\n");
+	if (display_to)
+		return 0;
+	return 1;
+}
+
 int proc_logfile(char *logfile)
 {
-	int fd, rc = -1;
-	size_t size;
-	struct bt_record *p_top, *p, *p_max;
+	int fd, rc = -1, tmp;
+	off_t size;
 	char *dir, *func_name;
+	struct proc_each_range_data data;
 
-	if (open_mmapfile(logfile, &fd, (void**)&p_top, &size) < 0)
+	if ((fd = u_open(logfile, &size)) < 0)
 		goto EXIT;
 
-	p = p_top;
 	func_name = basename(logfile);
 	dir = dirname(logfile);
 	if (is_exists_maps(dir, func_name)) {
-		if (parse_maps(add_range_to_name, dir, func_name) < 0)
+		if (parse_maps(&r2n_info, dir, func_name,
+			       PM_WITH_PT, PM_VERBOSE) <0)
 			goto EXIT;
 	}
-	if (parse_modules(add_range_to_name, dir) < 0)
+	if (parse_modules(&r2n_info, dir, PM_WITH_PT, PM_VERBOSE) < 0)
 		goto EXIT;
+	r2n_info.from_is_next = chk_from_is_next(dir);
+
+	if (search_sym) {
+		if (!get_symbol_addr(&r2n_info, search_sym, &search_addr)) {
+			fprintf(stderr, "symbol not found(%s).\n", search_sym);
+			goto EXIT;
+		}
+	}
 	printf("start\n");
 	if (verbose)
-		dump_r2n();
-	p_max = (struct bt_record*)((char*)p + size);
-	if (proc_block_records(p, p_max) < 0)
-		goto EXIT;
+		dump_r2n(&r2n_info);
+	if (search || display_from || search_warn) {
+		if (search) {
+			if (search_sym)
+				printf("search: %s(0x%08lx)",
+				       search_sym, search_addr);
+			else
+				printf("search: 0x%08lx", search_addr);
+			printf(" -B %lld -A %lld\n", before, after);
+		}
+		data.fd = fd;
+		data.i_rec_max = size / sizeof(struct bt_record);
+		data.i_start = -1;
+		data.i_stop = 0;
+		data.match_found = 0;
+		data.match_i_rec = -1;
+		do {
+			tmp = proc_search_range(data.i_stop, &data);
+			if (tmp < 0)
+				goto EXIT;
+		} while (tmp);
+	} else {
+		if (proc_block_records(fd, 0, size / sizeof(struct bt_record))
+		    < 0)
+			goto EXIT;
+	}
 	rc = 0;
 EXIT:
-	close_mmapfile(fd, p_top, size);
+	u_close(fd);
 	return rc;
 }
 
 void err_exit(void)
 {
 	free_ranges();
-	free_r2n();
+	free_r2n(&r2n_info);
 	exit(1);
 }
 
@@ -936,83 +692,131 @@ void usage(void)
 {
 	fprintf(stderr, "bt_execpath %s\n", BT_EXECPATH_VER);
 	fprintf(stderr, "    %s\n\n", COPYRIGHT);
-	fprintf(stderr, "bt_execpath [-s] [-a top:end [...]] [-d top:end [...]] [--usr|--ker|--all]\n");
-	fprintf(stderr, "            -f logfile\n");
+#if 0
+	fprintf(stderr, "bt_execpath [-s] [-a top:end] [-d top:end] [--usr|--ker|--all]\n");
+	fprintf(stderr, "            [-S aors|-F recnum] [-B n] [-A n|-T recnum] -f logfile\n");
+#else
+	fprintf(stderr, "bt_execpath [-s] [-a top:end] [-d top:end] [--usr|--ker|--all] -f logfile\n");
+#endif
 	fprintf(stderr, "  -s: output execution path summary\n");
 	fprintf(stderr, "  -a: add address range\n");
 	fprintf(stderr, "  -d: delete address range\n");
 	fprintf(stderr, "  --usr: alias of '-a 0:0xbfffffff'\n");
 	fprintf(stderr, "  --ker: alias of '-a 0xc0000000:0xffffffff'\n");
 	fprintf(stderr, "  --all: alias of '-a 0:0xffffffff'\n");
+#if 0
+	fprintf(stderr, "  -S: search address or symbol\n");
+	fprintf(stderr, "  -B: print n records before the record of -S or -F option\n");
+	fprintf(stderr, "  -A: print n records after the record of -S or -F option\n");
+	fprintf(stderr, "  -F: specify the start record number\n");
+	fprintf(stderr, "  -T: specify the end record number\n");
+#endif
 	fprintf(stderr, "  -f: logfile\n");
 }
 
 int main(int argc, char *argv[])
 {
-	int i;
-	char *logfile = NULL;
+	int opt_index;
+	char c, *logfile = NULL, *p_end;
 	unsigned long begin, end;
+	long tmp;
+	off_t tmp_index;
+	struct option long_options[] = {
+		{"usr", no_argument, NULL, 0},
+		{"ker", no_argument, NULL, 0},
+		{"all", no_argument, NULL, 0},
+	};
 
 	if (alloc_pid_range(ALL_PID) < 0)
 		err_exit();
-	for (i = 1; i < argc;) {
-		if (strcmp(argv[i], "-s") == 0) {
-			i++;
+	while ((c = getopt_long(argc, argv, "sa:d:S:F:B:A:T:f:Wv", long_options,
+				&opt_index)) != -1) {
+		switch (c) {
+		case 0:
+			switch (opt_index) {
+			case 0:
+				add_range(ALL_PID, 0, 0xbfffffff);
+				break;
+			case 1:
+				add_range(ALL_PID, 0xc0000000, 0xffffffff);
+				break;
+			case 2:
+				add_range(ALL_PID, 0, 0xffffffff);
+				break;
+			}
+			break;
+		case 's':
 			output_summary = 1;
-		} else if (strcmp(argv[i], "--usr") == 0) {
-			add_range(ALL_PID, 0, 0xbfffffff);
-			i++;
-		} else if (strcmp(argv[i], "--ker") == 0) {
-			add_range(ALL_PID, 0xc0000000, 0xffffffff);
-			i++;
-		} else if (strcmp(argv[i], "--all") == 0) {
-			add_range(ALL_PID, 0, 0xffffffff);
-			i++;
-		} else if (strcmp(argv[i], "-a") == 0) {
-			chk_next(argc, i);
-			i++;
-			while (i < argc && argv[i][0] != '-') {
-				if (!range2ulongs(argv[i], &begin, &end))
-					err_exit();
+			break;
+		case 'a':
+		case 'd':
+			if (!range2ulongs(optarg, &begin, &end))
+				err_exit();
+			if (c == 'a')
 				add_range(ALL_PID, begin, end);
-				i++;
-			}
-		} else if (strcmp(argv[i], "-d") == 0) {
-			chk_next(argc, i);
-			i++;
-			while (i < argc && argv[i][0] != '-') {
-				if (!range2ulongs(argv[i], &begin, &end))
-					err_exit();
-				printf("del_range\n");
+			else
 				del_range(ALL_PID, begin, end);
-				i++;
+			break;
+		case 'S':
+			tmp = strtoul(optarg, &p_end, 0);
+			if (*p_end != '\0' && *p_end != ' ')
+				search_sym = optarg;
+			search = 1;
+			search_addr = tmp;
+			break;
+		case 'B':
+		case 'A':
+		case 'F':
+		case 'T':
+			tmp_index = strtoull(optarg, &p_end, 0);
+			if (tmp_index < 0
+			    || (*p_end != '\0' && *p_end != ' ')) {
+				fprintf(stderr, "invalid value(%s).\n", optarg);
+				err_exit();
 			}
-		} else if (strcmp(argv[i], "-f") == 0) {
-			chk_next(argc, i);
-			i++;
-			logfile = argv[i];
-			i++;
-		} else if (strcmp(argv[i], "-v") == 0) {
+			switch (c) {
+			case 'B':
+				before = tmp_index;
+				break;
+			case 'A':
+				after = tmp_index;
+				break;
+			case 'F':
+				display_from = tmp_index;
+				break;
+			case 'T':
+				display_to = tmp_index;
+				break;
+			}
+			break;
+		case 'f':
+			logfile = optarg;
+			break;
+		case 'W':
+			search_warn = 1;
+			break;
+		case 'v':
 			verbose = 1;
-			i++;
-		} else {
+			break;
+		default:
 			usage();
 			err_exit();
 		}
 	}
 	//dump_ranges();	/* for debug */
-	if (!logfile) {
+	if (optind < argc ||
+	    !logfile || (search && display_from) || (after && display_to) ||
+	    (!search && !display_from && (after || display_to))) {
 		usage();
 		err_exit();
 	}
-	all_r2n = calloc(r2n_max, sizeof(struct range_to_name*));
-	if (!all_r2n) {
-		fprintf(stderr, "calloc failed.(%s)\n", strerror(errno));
-		err_exit();
-	}
+	if (search_warn)
+		after = 0;
+	r2n_info.all_r2n = xcalloc(r2n_info.r2n_max,
+				   sizeof(struct range_to_name*));
 	if (proc_logfile(logfile) < 0)
 		err_exit();
 	free_ranges();
-	free_r2n();
+	free_r2n(&r2n_info);
 	exit(0);
 }
